/*
 * rlm_eap_tnc.c    Handles that are called from eap
 *
 *   This software is Copyright (C) 2006-2008 FH Hannover
 *
 *   Portions of this code unrelated to FreeRADIUS are available
 *   separately under a commercial license.  If you require an
 *   implementation of EAP-TNC that is not under the GPLv2, please
 *   contact tnc@inform.fh-hannover.de for details.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 *   Modifications to integrate with FreeRADIUS configuration
 *   Copyright (C) 2007 Alan DeKok <aland@deployingradius.com>
 */

/*
 * EAP-TNC Packet with EAP Header, general structure
 *
 *  0                   1                   2                   3
 *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |      Code     |   Identifier  |            Length             |
 * |               |               |                               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |      Type     |  Flags  | Ver |          Data Length          |
 * |               |L M S R R| =1  |                               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |          Data Length          |           Data ...
 * |                               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

#include <freeradius-devel/autoconf.h>

#include <stdio.h>
#include <stdlib.h>

#include "eap_tnc.h"
#include <naatncs.h>
#include <freeradius-devel/rad_assert.h>
//#include <freeradius-devel/libradius.h>

#define VERSION "0.4.4"

/*
 * Keep configuration data in one struct.
 */
typedef struct rlm_eap_tnc_configuration {
	char *vlan_access;
	char *vlan_isolate;
} rlm_eap_tnc_configuration;

/*
 * define how the eap_tnc configuration must look like
 */
static CONF_PARSER module_config[] =
{
	{
		"vlan_access",
		PW_TYPE_STRING_PTR,
		offsetof(rlm_eap_tnc_configuration, vlan_access),
		NULL,
		NULL
	}, {
		"vlan_isolate",
		PW_TYPE_STRING_PTR,
		offsetof(rlm_eap_tnc_configuration,	vlan_isolate),
		NULL,
		NULL
	}, { NULL, -1, 0, NULL, NULL } /* end the list */
};

static TNC_ConnectionID calculateConnectionID(RADIUS_PACKET* radius_packet) {
	VALUE_PAIR *vp = NULL;
	uint32_t outMD5[4]; // 4 x 32 == 128 :-)
	uint32_t inMD5[3];
	uint32_t nas_port = 0;
	uint32_t nas_ip = 0;
	uint32_t nas_port_type = 0;
	TNC_ConnectionID ret = 0;


	if (radius_packet == NULL) {
		radlog(L_ERR,
				"rlm_eap_tnc: calculateConnectionID failed. radius_packet == NULL!");
		return 0;
	}

	for (vp = radius_packet->vps; vp; vp=vp->next) {
		switch (vp->attribute) {
		case PW_NAS_PORT:
			nas_port = vp->vp_integer;
			DEBUG("NAS scr port = %u\n", nas_port);
			break;
		case PW_NAS_IP_ADDRESS:
			nas_ip = vp->vp_ipaddr;
			DEBUG("NAS scr ip = %X\n", nas_ip);
			break;
		case PW_NAS_PORT_TYPE:
			nas_port_type = vp->vp_integer;
			DEBUG("NAS scr port type = %u\n", nas_port_type);
			break;
		}
	}

	// check that all attributes have been set
	// FIXME get values from parent request if possible. does not work as tunneled eap method atm
	if(nas_ip == 0 ||
	   nas_port_type == 0){
		radlog(L_ERR,
				"rlm_eap_tnc: not all attributes set for calculating connection ID.");
		return 0;
	}

	inMD5[0] = nas_ip;
	inMD5[1] = nas_port;
	inMD5[2] = nas_port_type;

	fr_md5_calc((uint8_t *)outMD5, (uint8_t *)inMD5, sizeof(uint32_t) * 3);

	ret = *((TNC_ConnectionID *)(void *)outMD5);
	return ret;
}

/*
 *	Detach the EAP-TNC module.
 */
static int tnc_detach(void *arg) {
	DEBUG("tnc_detach()");
	free(arg);
	return 0;
}

/*
 * This function is called to load the EAP_TNC module and "attaches"
 * it to FreeRADIUS.
 *
 * Basically, it parses the EAP_TNC configuration from
 * etc/raddb/eap.conf.
 */
static int tnc_attach(CONF_SECTION *cs, void **instance) {
	DEBUG("tnc_attach()");
	DEBUG("rlm_eap_tnc Version: %s", VERSION);
	// allocate memory for configuration data
	// this memory is freed within tnc_detach()
	rlm_eap_tnc_configuration *config;
	config = calloc(1, sizeof(*config));
	if (!config) {
		radlog(L_ERR, "rlm_eap_tnc:tnc_attach: calloc failed!");
		return -1;
	}

	// parse configuration
	if (cf_section_parse(cs, config, module_config) < 0) {
		radlog(L_ERR,
				"rlm_eap_tnc:tnc_attach: cf_section_parse failed! Detaching eap_tnc module.");
		tnc_detach(config);
		return -1;
	}

	// check if configuration is ok
	if (!config->vlan_access || !config->vlan_isolate) {
		radlog(L_ERR, "rlm_eap_tnc: Must set both vlan_access and vlan_isolate");
		tnc_detach(config);
		return -1;
	}

	*instance = config;
	return 0;
}

/*
 * This function is called when the first EAP_IDENTITY_RESPONSE message
 * was received.
 *
 * Initiates the EPA_TNC session by sending the first EAP_TNC_RESPONSE
 * to the peer. The packet has the Start-Bit set and contains no data.
 *
 *  0                   1                   2                   3
 *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |      Code     |   Identifier  |            Length             |
 * |               |               |                               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |      Type     |  Flags  | Ver |
 * |               |0 0 1 0 0|0 0 1|
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 * For this package, only 'Identifier' has to be set dynamically. Any
 * other information is static.
 */
static int tnc_initiate(void *type_data, EAP_HANDLER *handler) {

	TNC_ConnectionID conID = 0;
	REQUEST * request = NULL;

	// check if we run inside a secure EAP method.
	// FIXME check concrete outer EAP method
	if (!handler->request || !handler->request->parent) {
		DEBUG2("rlm_eap_tnc: EAP_TNC must only be used as an inner method within a protected tunneled EAP created by an outer EAP method.");
		request = handler->request;
		return 0;
	} else {
		request = handler->request->parent;
	}

	if (request->packet == NULL) {
		DEBUG2("rlm_eap_tnc: ERROR request->packet is NULL.");
		return 0;
	}

	DEBUG("tnc_initiate: %ld", handler->timestamp);


	//calculate connectionID
	conID = calculateConnectionID(request->packet);
	if(conID == 0){
		radlog(L_ERR, "rlm_eap_tnc:tnc_attach:Calculating connectionID failed.");
		return 0;
	}

	DEBUG2("TNC-INITIATE is creating now for connection ID %lX !", conID);

	/*
	 * create connection
	 * (uses a function from the NAA-TNCS-library)
	 * the presence of the library is checked via the configure-script
	 */
	TNC_Result result = connectionCreate(conID);

	// check for errors
	if (result != TNC_RESULT_SUCCESS) {
		radlog(L_ERR,
				"rlm_eap_tnc: tnc_authenticate error while calling NAA-TNCS createConnection");
		return 0;
	}

	// set connection ID
	handler->opaque = malloc(sizeof(TNC_ConnectionID));
	memcpy(handler->opaque, &conID, sizeof(TNC_ConnectionID));

	// build first EAP TNC request
	TNC_BufferReference eap_tnc_request = malloc(sizeof(unsigned char));
	if (eap_tnc_request == NULL) {
		radlog(L_ERR, "rlm_eap_tnc:tnc_initiate: malloc failed.");
		return 0;
	}
	*eap_tnc_request = SET_START(1);
	TNC_UInt32 eap_tnc_length = 1;
	type_data = type_data; /* suppress -Wunused */

	/*
	 *	Compose the EAP-TNC packet out of the data structure,
	 *	and free it.
	 */
	eaptnc_compose(handler, eap_tnc_request, eap_tnc_length, PW_EAP_REQUEST);

	/*
	 *	We don't need to authorize the user at this point.
	 *
	 *	We also don't need to keep the challenge, as it's
	 *	stored in 'handler->eap_ds', which will be given back
	 *	to us...
	 */
	handler->stage = AUTHENTICATE;

	return 1;
}

static void setVlanAttribute(rlm_eap_tnc_configuration *config,
		EAP_HANDLER *handler, VlanAccessMode mode) {
	VALUE_PAIR *vp;
	char *vlanNumber = NULL;
	switch (mode) {
	case VLAN_ISOLATE:
		vlanNumber = config->vlan_isolate;
		vp = pairfind(handler->request->config_items, PW_TNC_VLAN_ISOLATE);
		if (vp)
			vlanNumber = vp->vp_strvalue;
		break;
	case VLAN_ACCESS:
		vlanNumber = config->vlan_access;
		vp = pairfind(handler->request->config_items, PW_TNC_VLAN_ACCESS);
		if (vp)
			vlanNumber = vp->vp_strvalue;
		break;

	default:
		DEBUG2("  rlm_eap_tnc: Internal error.  Not setting vlan number");
		return;
	}

	if (!handler->request->parent) {
		DEBUG2("  rlm_eap_tnc: EAP_TNC is used without an outer method.");

		pairadd(&handler->request->reply->vps, pairmake("Tunnel-Type", "VLAN",
					T_OP_SET));
			// FIXME: RFC 3580 says '802' instead of 'IEEE-802'
			pairadd(&handler->request->reply->vps, pairmake("Tunnel-Medium-Type",
					"IEEE-802", T_OP_SET));

			pairadd(&handler->request->reply->vps, pairmake("Tunnel-Private-Group-ID",
					vlanNumber, T_OP_SET));
	} else {
		DEBUG2("  rlm_eap_tnc: EAP_TNC is used in a protected tunneled EAP");

		pairadd(&handler->request->parent->reply->vps, pairmake("Tunnel-Type", "VLAN",
					T_OP_SET));
			// FIXME: RFC 3580 says '802' instead of 'IEEE-802'
			pairadd(&handler->request->parent->reply->vps, pairmake("Tunnel-Medium-Type",
					"IEEE-802", T_OP_SET));

			pairadd(&handler->request->parent->reply->vps, pairmake("Tunnel-Private-Group-ID",
					vlanNumber, T_OP_SET));
	}
}

/**
 * This function is called when a EAP_TNC_RESPONSE was received.
 * It basically forwards the EAP_TNC data to NAA-TNCS and forms
 * and appropriate EAP_RESPONSE. Furthermore, it sets the VlanID
 * based on the TNC_ConnectionState determined by NAA-TNCS.
 *
 * @param type_arg The configuration data
 * @param handler The EAP_HANDLER
 * @return True, if successfully, else false.
 */
static int tnc_authenticate(void *type_arg, EAP_HANDLER *handler) {

	rad_assert(handler->request != NULL); // check that request has been sent previously
	rad_assert(handler->stage == AUTHENTICATE); // check if initiate has been called

	if (handler == NULL) {
		radlog(L_ERR,
				"rlm_eap_tnc: tnc_authenticate invalid parameters: handler == NULL");
		return 0;
	}
	if (handler->eap_ds == NULL) {
		radlog(L_ERR,
				"rlm_eap_tnc: tnc_authenticate invalid parameters: handler->eap_ds == NULL");
		return 0;
	}
	if (handler->eap_ds->response == NULL) {
		radlog(
				L_ERR,
				"rlm_eap_tnc: tnc_authenticate invalid parameters: handler->eap_ds->resonse == NULL");
		return 0;
	}
	if (handler->eap_ds->response->type.type != PW_EAP_TNC
			|| handler->eap_ds->response->type.length < 1
			|| handler->eap_ds->response->type.data == NULL) {
		radlog(
				L_ERR,
				"rlm_eap_tnc: tnc_authenticate invalid parameters: handler->eap_ds->response->type.type == %X, ->type.length == %u, ->type.data == %p",
				handler->eap_ds->response->type.type,
				handler->eap_ds->response->type.length,
				handler->eap_ds->response->type.data);
		return 0;
	}

	// get connection ID
	TNC_ConnectionID conID = *((TNC_ConnectionID *) (handler->opaque));

	DEBUG2("TNC-AUTHENTICATE is starting now for connection ID %lX !", conID);

	// pass EAP_TNC data to NAA-TNCS and get answer data
	TNC_BufferReference output = NULL;
	TNC_UInt32 outputLength = 0;
	TNC_ConnectionState connectionState = TNC_CONNECTION_STATE_CREATE;

	/*
	 * exchanges the tnc message
	 * (uses a function from the NAA-TNCS-library)
	 * the presence of the library is checked via the configure-script
	 */
	TNC_Result result = exchangeTNCMessage(conID, handler->eap_ds->response->type.data,
			handler->eap_ds->response->type.length, &output, &outputLength,
			&connectionState);

	// check for errors
	if (result != TNC_RESULT_SUCCESS) {
		radlog(L_ERR,
				"rlm_eap_tnc: tnc_authenticate error while calling NAA-TNCS");
		return 0;
	}

	// configuration data parsed in attach()
	rlm_eap_tnc_configuration *config = type_arg;
	// output contains now the answer from NAA-TNCS
	uint8_t eapCode = 0;
	// determine eapCode for request
	switch (connectionState) {
	case TNC_CONNECTION_STATE_HANDSHAKE:
		eapCode = PW_EAP_REQUEST;
		break;
	case TNC_CONNECTION_STATE_ACCESS_NONE:
		eapCode = PW_EAP_FAILURE;
		break;
	case TNC_CONNECTION_STATE_ACCESS_ALLOWED:
		eapCode = PW_EAP_SUCCESS;
		setVlanAttribute(config, handler, VLAN_ACCESS);
		break;
	case TNC_CONNECTION_STATE_ACCESS_ISOLATED:
		eapCode = PW_EAP_SUCCESS;
		setVlanAttribute(config, handler, VLAN_ISOLATE);
		break;
	default:
		radlog(L_ERR,
				"rlm_eap_tnc: tnc_authenticate invalid TNC_CONNECTION_STATE.");
		return 0;
	}

	// form EAP_REQUEST
	if (!eaptnc_compose(handler, output, outputLength, eapCode)) {
		radlog(L_ERR,
				"rlm_eap_tnc: tnc_authenticate error while forming EAP_REQUEST.");
		return 0;
	}

	// FIXME: Why is that needed?
	handler->stage = AUTHENTICATE;

	return 1;
}

/*
 *	The module name should be the only globally exported symbol.
 *	That is, everything else should be 'static'.
 */
EAP_TYPE rlm_eap_tnc = {
		"eap_tnc",
		tnc_attach, /* attach */
		tnc_initiate, /* Start the initial request */
		NULL, /* authorization */
		tnc_authenticate, /* authentication */
		tnc_detach /* detach */
};
